/*
   ZoranIO.c

   ZORAN ZR36120 module driver.
   Copyright (C) 1997  Vincent Labie (try ivsparis@pratique.fr with
				      for Vincent Labie in the suject)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	"ZoranIO.h"
#include	"SAA7110.h"

#define		ZFS_NOTFILLED			0x01234567

#define		ZORAN_50HZ_MAX_WIDTH		768
#define		ZORAN_50HZ_MAX_HEIGHT		576
#define		ZORAN_50HZ_OFFSET_HEIGHT	151

#define		ZORAN_60HZ_MAX_WIDTH		640
#define		ZORAN_60HZ_MAX_HEIGHT		480
#define		ZORAN_60HZ_OFFSET_HEIGHT	125

#define		ZORAN_MAX_BPP			32
#define		ZORAN_SCREEN			3

int		GenPurDir	=ZORAN_GENPURDIR;
int		GenPurIO	=ZORAN_GENPURDIR;
int		GIRQ		=ZORAN_GIRQ;
int		ExtFI		=ZORAN_EXTFI;
int		TopField	=ZORAN_TOPFIELD;
int		HSPol		=ZORAN_HSPOL;
int		VSPol		=ZORAN_VSPOL;

extern	ZORAN *		TabZoran[NB_MAX_ZORAN_CARD];

	void	FreeFrameBuffer(ZORAN * pZoran);
	BOOL	AllocFrameBuffer(ZORAN * pZoran,int NbSlot,u32 FrameSize);

/**********************************************************************************************************************/
	BOOL	ZoranReset(const ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;
	pZR->SysPCIControl.SoftReset		=0;	// reset ZR36120
	return(TRUE);
}

/**********************************************************************************************************************/
#define  TRITON_PCON	           0x50 
#define  TRITON_BUS_CONCURRENCY   (1<<0)
#define  TRITON_STREAMING	  (1<<1)
#define  TRITON_WRITE_BURST	  (1<<2)
#define  TRITON_PEER_CONCURRENCY  (1<<3)

/*#ifdef	DONT_TOUCH_MY_CHIPSET
 #comment Your chipset setings will not be changed. very bad frame rate can result. look at the DONT_TOUCH_MY_CHIPSET define
#else	//DONT_TOUCH_MY_CHIPSET
 #comment WARNING: Your chipset setings may be changed when using this driver. Be carefull!
#endif	//DONT_TOUCH_MY_CHIPSET*/

static	BOOL	ZoranFind437FX(void)
{
	u8			IndexPCI,Bus,dfn;
	u8			Byte,Byte0;
	int			ChipSet=0;

        for (IndexPCI=0;IndexPCI<8;IndexPCI++)
		{
		if (pcibios_find_device (PCI_VENDOR_ID_INTEL,PCI_DEVICE_ID_INTEL_82437,
			IndexPCI,&Bus,&dfn)==PCIBIOS_SUCCESSFUL)
				{ ChipSet=PCI_DEVICE_ID_INTEL_82437;
				ZRDBG(("Triton 82437FX found... too bad...\n")); break; }
		if (pcibios_find_device (PCI_VENDOR_ID_INTEL,PCI_DEVICE_ID_INTEL_82439,
			IndexPCI,&Bus,&dfn)==PCIBIOS_SUCCESSFUL)
				{ ChipSet=PCI_DEVICE_ID_INTEL_82439;
				ZRDBG(("Triton 82439HX found.\n")); break; }
		if (pcibios_find_device (PCI_VENDOR_ID_INTEL,PCI_DEVICE_ID_INTEL_82437VX,
			IndexPCI,&Bus,&dfn)==PCIBIOS_SUCCESSFUL)
				{ ChipSet=PCI_DEVICE_ID_INTEL_82437VX;
				ZRDBG(("Triton 82439VX found.\n")); break; }
		}
	switch(ChipSet)
		{
		case PCI_DEVICE_ID_INTEL_82437:
			// stolen from Ralph Metzler's bttv driver
			pcibios_read_config_byte(Bus,dfn,TRITON_PCON,&Byte0);
			Byte=Byte0;

			// REALLY needed, so is default
			/* 430FX (Triton I) freezes with bus concurrency on -> switch it off */
			if((Byte&TRITON_BUS_CONCURRENCY)==0)
				{ ERR(("Zoran: 82437FX: disabling bus concurrency\n"));
				Byte|=TRITON_BUS_CONCURRENCY; }

			/* still freezes on other boards -> switch off even more */
			if((Byte&TRITON_PEER_CONCURRENCY)!=0)
				{ ERR(("Zoran: 82437FX: disabling peer concurrency\n"));
				Byte&=~TRITON_PEER_CONCURRENCY; }
#ifndef	DONT_TOUCH_MY_CHIPSET
			// from me:
			// must be set to have more than .2 fps with a trio64/1Mb!
			if(Byte&TRITON_STREAMING)
				{ ERR(("Zoran: 82437FX: enabling streaming\n"));
				Byte&=~TRITON_STREAMING; }
#endif	//DONT_TOUCH_MY_CHIPSET
			if (Byte!=Byte0)
				{ pcibios_write_config_byte(Bus,dfn,TRITON_PCON,Byte); 
				ERR(("Zoran: 82437FX: PCON changed to: 0x%02X\n",Byte)); }
			return(TRUE);
		case PCI_DEVICE_ID_INTEL_82437VX:
		case PCI_DEVICE_ID_INTEL_82439:
			pcibios_read_config_byte(Bus,dfn,TRITON_PCON,&Byte0);
			Byte=Byte0;

#ifndef	DONT_TOUCH_MY_CHIPSET
			// from me:
			// useful, but not requered, so not default
			// from test-and-try
			if((Byte&TRITON_BUS_CONCURRENCY)==0)
				{ ERR(("Zoran: 82437VX/82439HX: disabling bus concurrency\n"));
				Byte|=TRITON_BUS_CONCURRENCY; }

			if((Byte&TRITON_PEER_CONCURRENCY)==0)
				{ ERR(("Zoran: 82437VX/82439HX: enabling peer concurrency\n"));
				Byte|=TRITON_PEER_CONCURRENCY; }

			// from me:
			// needed by HX/VX mother board with VERY cheap BIOS :[
			// must be set to had more than .5 fps with a ET6000!
			if(Byte&TRITON_STREAMING)
				{ ERR(("Zoran: 82437VX/82439HX: enabling streaming\n"));
				Byte&=~TRITON_STREAMING; }
#endif	//DONT_TOUCH_MY_CHIPSET
			if (Byte!=Byte0)
				{ pcibios_write_config_byte(Bus,dfn,TRITON_PCON,Byte); 
				ERR(("Zoran: 82437VX/82439HX: PCON changed to: 0x%02X\n",Byte)); }
			return(FALSE);
		default:
			return(FALSE);
		}
}

/**********************************************************************************************************************/
	BOOL	ZoranUnreset(const ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;

	pZR->SysPCIControl.SoftReset				=1;		// free ZR36120
	if (ZoranFind437FX())	pZR->VideoControl.TriCom	=1;
	else			pZR->VideoControl.TriCom	=0;
	pZR->SysPCIControl.GenPurDir				=GenPurDir;	// free SAA7110
	pZR->GenPurpose.GenPurIO				=GenPurIO;
//	pZR->VideoControl.MinPix				=7;

	return(SAA7110Init(pZoran));
}

/**********************************************************************************************************************/
	int	ZoranCapParams(ZORAN * pZoran,ZR_CAP_PARAMS * pZRCapParams)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;
	int			FrameSize;

	if (pZR->VideoControl.VidEn)
		{ ERR(("Zoran%d: try to modify CapParams while running\n",pZoran->IndexPCI)); return(-EBUSY); }

	ZRDBG(("pZRCapParams->CapWidth=%i\n",		pZRCapParams->CapWidth));
	ZRDBG(("pZRCapParams->CapHeight=%i\n",		pZRCapParams->CapHeight));
	ZRDBG(("pZRCapParams->CapScanLineSize=%i\n",	pZRCapParams->CapScanLineSize));
	ZRDBG(("pZRCapParams->ZRReduceMode=%i\n",	pZRCapParams->ZRReduceMode));
	ZRDBG(("pZRCapParams->ZRPixel=%i\n",		pZRCapParams->ZRPixel));
	ZRDBG(("pZRCapParams->ZREndian=%i\n",		pZRCapParams->ZREndian));
	ZRDBG(("pZRCapParams->ZRField=%i\n",		pZRCapParams->ZRField));
	ZRDBG(("pZRCapParams->ZRHFilter=%i\n",		pZRCapParams->ZRHFilter));
	ZRDBG(("pZRCapParams->ZRErrDif=%i\n",		pZRCapParams->ZRErrDif));
	ZRDBG(("pZRCapParams->NbSlot=%i\n",		pZRCapParams->NbSlot));

	if ((pZRCapParams->CapWidth<=0)||(pZRCapParams->CapWidth>ZORAN_50HZ_MAX_WIDTH))
		{ ERR(("Zoran%d: invalide CapWidth=%i\n",
			pZoran->IndexPCI,pZRCapParams->CapWidth)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->CapWidth ok\n"));
	if ((pZRCapParams->CapHeight<=0)||(pZRCapParams->CapHeight>ZORAN_50HZ_MAX_HEIGHT))
		{ ERR(("Zoran%d: invalide CapHeight=%i\n",
			pZoran->IndexPCI,pZRCapParams->CapHeight)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->CapHeight ok\n"));
	if (pZRCapParams->CapScanLineSize&0x03)
		{ ERR(("Zoran%d: invalide CapScanLineSize=%i, not DWORD alined\n",
			pZoran->IndexPCI,pZRCapParams->CapScanLineSize)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->CapScanLineSize ok\n"));
	if ((pZRCapParams->ZRReduceMode<=ZRM_TOOSMALL)||(pZRCapParams->ZRReduceMode>=ZRM_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZRReduceMode=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZRReduceMode)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZRReduceMode ok\n"));
	if ((pZRCapParams->ZRPixel<=ZPX_TOOSMALL)||(pZRCapParams->ZRPixel>=ZPX_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZRPixel=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZRPixel)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZRPixel ok\n"));
	if ((pZRCapParams->ZREndian<=ZEN_TOOSMALL)||(pZRCapParams->ZREndian>=ZEN_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZREndian=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZREndian)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZREndian ok\n"));
	if ((pZRCapParams->ZRField<=ZFI_TOOSMALL)||(pZRCapParams->ZRField>=ZFI_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZRField=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZRField)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZRField ok\n"));
	if ((pZRCapParams->ZRHFilter<=ZHF_TOOSMALL)||(pZRCapParams->ZRHFilter>=ZHF_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZRHFilter=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZRHFilter)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZRHFilter ok\n"));
	if ((pZRCapParams->ZRErrDif<=ZED_TOOSMALL)||(pZRCapParams->ZRErrDif>=ZED_TOOLARGE))
		{ ERR(("Zoran%d: invalide ZRErrDif=%i\n",
			pZoran->IndexPCI,pZRCapParams->ZRErrDif)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->ZRErrDif ok\n"));
	if ((pZRCapParams->NbSlot<=0)||(pZRCapParams->NbSlot>NB_MAX_ZORAN_SLOT))
		{ ERR(("Zoran%d: invalide NbSlot=%i\n",
			pZoran->IndexPCI,pZRCapParams->NbSlot)); return(-EINVAL); }
	ZRDBG(("pZRCapParams->NbSlot ok\n"));

	pZoran->CapWidth			=pZRCapParams->CapWidth;
	pZoran->CapHeight			=pZRCapParams->CapHeight;
	pZoran->bCrop				=pZRCapParams->ZRReduceMode==ZRM_CROP;
	switch(pZRCapParams->ZRPixel)
		{
		case ZPX_YUV422:		pZR->OutVideoControl.YUV2RGB	=0;
						pZoran->CapPixelBytes=2;		break;
		case ZPX_RGB888_PACK32:		pZR->OutVideoControl.YUV2RGB	=1;
						pZR->OutVideoControl.Pack24	=0;
						pZoran->CapPixelBytes=4;		break;
		case ZPX_RGB888_PACK24:		pZR->OutVideoControl.YUV2RGB	=1;
						pZR->OutVideoControl.Pack24	=1;
						pZoran->CapPixelBytes=3;		break;
		case ZPX_RGB565:		pZR->OutVideoControl.YUV2RGB	=2;
						pZoran->CapPixelBytes=2;		break;
		case ZPX_RGB555:		pZR->OutVideoControl.YUV2RGB	=3;
						pZoran->CapPixelBytes=2;		break;
		case ZPX_TOOSMALL:
		case ZPX_TOOLARGE:
			break;
		}
	pZoran->ScanLineSize			=pZoran->CapWidth*pZoran->CapPixelBytes;
	if (pZoran->ScanLineSize&0x03)
		{ ERR(("Zoran%d: real ScanLineSize=CapWidth*CapPixelBytes is not DWORD alined.\n,",
			pZoran->IndexPCI)); return(-EINVAL); }

	if (pZRCapParams->CapScanLineSize<=0)
		pZRCapParams->CapScanLineSize	=pZoran->ScanLineSize;
	pZoran->CapScanLineSize			=pZRCapParams->CapScanLineSize;
	if (pZoran->CapScanLineSize&0x03)
		{ ERR(("Zoran%d: real CapScanLineSize=%i is not DWORD alined.\n,",
			pZoran->IndexPCI,pZoran->CapScanLineSize)); return(-EINVAL); }

	if (pZoran->ScanLineSize>pZRCapParams->CapScanLineSize)
		{ ERR(("Zoran%d: CapScanLineSize=%i is less than real ScanLineSize=%i.\n",
			pZoran->IndexPCI,pZRCapParams->CapScanLineSize,pZRCapParams->CapScanLineSize));
		return(-EINVAL); }

	pZR->OutVideoControl.LittleEndian	=(pZRCapParams->ZREndian==ZEN_LITTLEENDIAN);
	pZR->OutVideoControl.DupFld		=(pZRCapParams->ZRField==ZFI_DUPLICATE);
	pZR->OutVideoControl.HFilter		=(int)pZRCapParams->ZRHFilter;
	pZR->OutVideoControl.ErrDif		=(pZRCapParams->ZRErrDif==ZED_ACTIVE);

	FreeFrameBuffer(pZoran);
	FrameSize=pZoran->CapScanLineSize*(pZoran->CapHeight);

	if (AllocFrameBuffer(pZoran,pZRCapParams->NbSlot,FrameSize)==FALSE)
		return(-ENOMEM);
	memcpy(pZRCapParams->TabSlotOffset,pZoran->TabSlotOffset,sizeof(pZRCapParams->TabSlotOffset));
	return(0);
}

/**********************************************************************************************************************/
	BOOL	ZoranSetMode(ZORAN * pZoran,int Command)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;
	int			hcrop1,hcrop2,HorDcm;
	int			vcrop1,vcrop2,VerDcm;
	int			ZoranMaxWidth,ZoranMaxHeight,ZoranOffsetHeight;

	if (pZoran->pFrameBuffer==NULL) return(-ENOMEM);
	if (((Command==ZORAN_CAPMODE_3CONTINOUSHACK)||(Command==ZORAN_CAPMODE_3CONTINOUS))&&
	    (pZoran->NbSlot<3)) return(-ENOMEM);
	switch(pZoran->VideoStd)
		{
		case ZORAN_NTSC:
		case ZORAN_BW60HZ:
			ZoranMaxWidth		=ZORAN_60HZ_MAX_WIDTH;
			ZoranMaxHeight		=ZORAN_60HZ_MAX_HEIGHT;
			ZoranOffsetHeight	=ZORAN_60HZ_OFFSET_HEIGHT;
			break;
		case ZORAN_PAL:
		case ZORAN_SECAM:
		case ZORAN_BW50HZ:
			ZoranMaxWidth		=ZORAN_50HZ_MAX_WIDTH;
			ZoranMaxHeight		=ZORAN_50HZ_MAX_HEIGHT;
			ZoranOffsetHeight	=ZORAN_50HZ_OFFSET_HEIGHT;
			break;
		default:
			ERR(("Zoran%d: SetMode with no standart.\n",pZoran->IndexPCI));
			return(-EIO);
		}
	if (pZoran->CapWidth>ZoranMaxWidth)
		{ ERR(("Zoran%d: SetMode CapWidth=%i > ZoranMaxWidth=%i.\n",pZoran->IndexPCI,
			pZoran->CapWidth,ZoranMaxWidth));
		return(-EINVAL); }

	if (pZoran->CapHeight>ZoranMaxHeight)
		{ ERR(("Zoran%d: SetMode CapHeight=%i > ZoranMaxHeight=%i.\n",pZoran->IndexPCI,
			pZoran->CapHeight,ZoranMaxHeight));
		return(-EINVAL); }

	pZR->IRQControl.InPinEn				=0;
	pZR->VideoControl.VidEn				=0;

	pZoran->CapSlot		=0;
	pZoran->UserSlot	=2;
	memset(pZoran->TabSlotFilled,0,sizeof(pZoran->TabSlotFilled));

	if (Command==ZORAN_CAPMODE_STOP) { pZoran->ZRCapMode=ZCM_STOP; return(0); }

	pZR->AdrTopVideo				=(u32)pZoran->pFrameBuffer+
								pZoran->TabSlotOffset[pZoran->CapSlot];
	pZR->AdrBottomVideo				=(u32)pZoran->pFrameBuffer+
								pZoran->TabSlotOffset[pZoran->CapSlot]+
								pZoran->CapScanLineSize;

	pZR->VideoControl.VidWinWid			=pZoran->CapWidth;
	if (pZoran->CapHeight<=ZoranMaxHeight/2)
		{
		pZR->VideoControl.VidWinHt		=pZoran->CapHeight;
		pZR->StrideGrabControl.DispStride	=pZoran->CapScanLineSize-pZoran->ScanLineSize;
		pZR->OutVideoControl.DispMod		=1;
		}
	else	{
		pZR->VideoControl.VidWinHt		=pZoran->CapHeight/2;
		pZR->StrideGrabControl.DispStride	=pZoran->CapScanLineSize*2-pZoran->ScanLineSize;
		pZR->OutVideoControl.DispMod		=0;
		}
	ZRDBG(("VidWinWid=%i, VidWinHt=%i, DispMode=%i\n",
		pZR->VideoControl.VidWinWid,pZR->VideoControl.VidWinHt,pZR->OutVideoControl.DispMod));

	if (pZoran->bCrop)
		{
		hcrop1		=(ZoranMaxWidth-pZR->VideoControl.VidWinWid)/2;
		hcrop2		=ZoranMaxWidth-pZR->VideoControl.VidWinWid-hcrop1;
		HorDcm		=0;
		ZRDBG(("Crop mode: hcrop1=%i, hcrop2=%i\n",hcrop1,hcrop2));
		}
	else	{
		int	Wa	=ZoranMaxWidth;
		int	X	=(int)((float)(pZR->VideoControl.VidWinWid*64)/(float)Wa+.999999999999);
		int	We	=(int)((float)(pZR->VideoControl.VidWinWid*64)/(float)X);
		hcrop1		=2*(int)((float)(Wa-We)/4.0);
		hcrop2		=Wa-We-hcrop1;
		HorDcm		=64-X;
		ZRDBG(("Scale mode: Wa=%i, X=%i, We=%i,\n",Wa,X,We));
		ZRDBG(("            hcrop1=%i, hcrop2=%i\n",hcrop1,hcrop2));
		}
	pZR->OutVideoControl.HorDcm			=HorDcm;
	pZR->HVideoControl.HStart			=ZoranOffsetHeight+hcrop1;
	pZR->HVideoControl.HEnd				=ZoranOffsetHeight+ZoranMaxWidth-1-hcrop2;
	ZRDBG(("HorDcm=%i, HStart=%i, HEnd=%i\n",pZR->OutVideoControl.HorDcm,
		pZR->HVideoControl.HStart,pZR->HVideoControl.HEnd));

	if (pZoran->bCrop)
		{
		vcrop1		=(ZoranMaxHeight/2-pZR->VideoControl.VidWinHt)/2;
		vcrop2		=ZoranMaxHeight/2-pZR->VideoControl.VidWinHt-vcrop1;
		VerDcm		=0;
		ZRDBG(("Crop mode: vcrop1=%i, vcrop2=%i\n",vcrop1,vcrop2));
		}
	else	{
		int	Ha	=ZoranMaxHeight/2;
		int	Y	=(int)((float)(pZR->VideoControl.VidWinHt*64)/(float)Ha+.9999999999999);
		int	He	=(int)((float)(pZR->VideoControl.VidWinHt*64)/(float)Y);
		vcrop1		=2*(int)((float)(Ha-He)/4.0);
		vcrop2		=Ha-He-vcrop1;
		VerDcm		=64-Y;
		ZRDBG(("Scale mode: Ha=%i, Y=%i, He=%i,\n",Ha,Y,He));
		ZRDBG(("            vcrop1=%i, vcrop2=%i\n",vcrop1,vcrop2));
		}
	pZR->OutVideoControl.VerDcm			=VerDcm;
	pZR->VVideoControl.VStart			=16+vcrop1;
	pZR->VVideoControl.VEnd				=16+(ZoranMaxHeight/2)-1-vcrop2;
	ZRDBG(("VerDcm=%i, VStart=%i, VEnd=%i\n", pZR->OutVideoControl.VerDcm,
		pZR->VVideoControl.VStart,pZR->VVideoControl.VEnd));

	pZR->OutVideoControl.ExtFI			=ExtFI;
	pZR->OutVideoControl.TopField			=TopField;
	pZR->HVideoControl.HSPol			=HSPol;
	pZR->VVideoControl.VSPol			=VSPol;

	if (GIRQ==0)	{ pZR->IRQControl.GIRQ0En	=1;	//enable
			pZR->IRQStatus.GIRQ0		=1; }	//clear
	else		{ pZR->IRQControl.GIRQ1En	=1;	//enable
			pZR->IRQStatus.GIRQ1		=1; }	//clear
	pZR->IRQControl.InPinEn				=1;
	switch(Command)
		{
		case ZORAN_CAPMODE_CONTINOUS:
			pZoran->ZRCapMode=ZCM_CONTINOUS;
			pZR->StrideGrabControl.SnapShot	=0;
			pZR->StrideGrabControl.FrameGrab=0;
			break;
		case ZORAN_CAPMODE_3CONTINOUSHACK:
			pZoran->ZRCapMode=ZCM_3CONTINOUSHACK;
			*(u32*)((u8*)pZoran->pFrameBuffer+
				pZoran->TabSlotOffset[pZoran->CapSlot]+
				pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+0))=ZFS_NOTFILLED;
			*(u32*)((u8*)pZoran->pFrameBuffer+
				pZoran->TabSlotOffset[pZoran->CapSlot]+
				pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+1))=ZFS_NOTFILLED;
			pZR->StrideGrabControl.SnapShot	=0;
			pZR->StrideGrabControl.FrameGrab=0;
			break;
		case ZORAN_CAPMODE_3CONTINOUS:
			pZoran->ZRCapMode=ZCM_3CONTINOUS;
			pZR->StrideGrabControl.SnapShot	=1;
			pZR->StrideGrabControl.FrameGrab=1;
			break;
		case ZORAN_CAPMODE_GRAB:
			pZoran->ZRCapMode=ZCM_GRAB;
			pZR->StrideGrabControl.SnapShot	=1;
			pZR->StrideGrabControl.FrameGrab=1;
			break;
		}
	pZR->VideoControl.VidEn				=1; // let's go
	return(0);
}

/**********************************************************************************************************************/
	int	ZoranCapIsGrab(ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;
	if (pZoran->ZRCapMode!=ZCM_GRAB)
		{ ERR(("Zoran%d IS_GRAB ioctl error: not in GRABMODE.\n",pZoran->IndexPCI)); return(-EINVAL); }
	return(pZR->StrideGrabControl.FrameGrab==0);
}

/**********************************************************************************************************************/
	int	ZoranCapNewGrab(ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;
	if (pZoran->ZRCapMode!=ZCM_GRAB)
		{ ERR(("Zoran%d NEW_GRAB ioctl error: not in GRABMODE.\n",pZoran->IndexPCI)); return(-EINVAL); }
	if (pZR->StrideGrabControl.FrameGrab==1)  { return(-EBUSY); }
	pZR->StrideGrabControl.FrameGrab=1;
	return(0);
}

/**********************************************************************************************************************/
static const u8	TabNextUserSlot[3][3]= { {1,2,1}, {2,2,0} , {1,0,0} };
	int	ZoranCapNewUser(ZORAN * pZoran)
{
	int			Error		=0;
	int			PrevCapSlot	=0;
	int			PrevUserSlot	=0;
	int			NextUserSlot;
	unsigned long		flags;

	if ((pZoran->ZRCapMode!=ZCM_3CONTINOUSHACK)&&(pZoran->ZRCapMode!=ZCM_3CONTINOUS))
		{ ERR(("Zoran%d NEW_USER ioctl error: not in 3CONTINOUS mode.\n",pZoran->IndexPCI)); return(-EINVAL); }

	if ((pZoran->CapSlot<0)||(pZoran->CapSlot>2))
		{ Error|=1; PrevCapSlot=pZoran->CapSlot; pZoran->CapSlot=0; }	//??????????? gasp!
	if ((pZoran->UserSlot<0)||(pZoran->UserSlot>2))
		{ Error|=2; PrevUserSlot=pZoran->UserSlot; pZoran->UserSlot=0; }//??????????? gasp!

	save_flags(flags); cli();
	NextUserSlot=TabNextUserSlot[pZoran->CapSlot][pZoran->UserSlot];
	if (pZoran->TabSlotFilled[NextUserSlot]==0) NextUserSlot=4;
	else	{ pZoran->TabSlotFilled[pZoran->UserSlot]=0;			// free old user slot
		pZoran->UserSlot=NextUserSlot; }				// set new user slot
	restore_flags(flags);

	if (Error&1) ERR(("Zoran%d cap slot err: %i.\n",pZoran->IndexPCI,PrevCapSlot));
	if (Error&2) ERR(("Zoran%d user slot err: %i.\n",pZoran->IndexPCI,PrevUserSlot));

	if (NextUserSlot==4) schedule(); 
	return(NextUserSlot);
}

/**********************************************************************************************************************/
static const u8	TabNextCapSlot[3][3]= { {1,2,1}, {2,2,0} , {1,0,0} };
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/

static inline	BOOL		ZoranIrq3ContinousHackIsFrameOK(ZORAN * pZoran)
{
	u32			ZFrameState;

	ZFrameState	=*(u32*)((u8*)pZoran->pFrameBuffer+
				pZoran->TabSlotOffset[pZoran->CapSlot]+
				pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+0));
	if (ZFrameState==ZFS_NOTFILLED)
		{ /*printk("field 1 not filled\n");*/ return(FALSE); }

	ZFrameState	=*(u32*)((u8*)pZoran->pFrameBuffer+
				pZoran->TabSlotOffset[pZoran->CapSlot]+
				pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+1));
	if (ZFrameState==ZFS_NOTFILLED)
		{ /*printk("field 2 not filled\n");*/ return(FALSE); }
	return(TRUE);
}

/**********************************************************************************************************************/
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/

static inline	BOOL		ZoranIrq3ContinousIsFrameOK(ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;

	if (pZR->StrideGrabControl.FrameGrab!=0) return(FALSE);
    	return(TRUE);
}

/**********************************************************************************************************************/
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/

static inline	void		ZoranIrq3ContinousSetNewFrame(ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;

	if ((pZoran->CapSlot<0)||(pZoran->CapSlot>2))
		pZoran->CapSlot=0;	//??????????? gasp!
	else	pZoran->TabSlotFilled[pZoran->CapSlot]=TRUE;
	if ((pZoran->UserSlot<0)||(pZoran->UserSlot>2))
		pZoran->UserSlot=0;	//??????????? gasp!
	pZoran->CapSlot=TabNextCapSlot[pZoran->CapSlot][pZoran->UserSlot];

	pZR->VideoControl.VidEn			=0;
	udelay(1);

	pZR->AdrTopVideo	=(u32)pZoran->pFrameBuffer+
					pZoran->TabSlotOffset[pZoran->CapSlot];
	pZR->AdrBottomVideo	=(u32)pZoran->pFrameBuffer+
					pZoran->TabSlotOffset[pZoran->CapSlot]+
					pZoran->CapScanLineSize;

	pZR->VideoControl.VidEn			=1;
}

/**********************************************************************************************************************/
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/

static inline	void		ZoranIrq3ContinousHackSetupNewFrame(ZORAN * pZoran)
{
	*(u32*)((u8*)pZoran->pFrameBuffer+
		pZoran->TabSlotOffset[pZoran->CapSlot]+
		pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+0))=ZFS_NOTFILLED;
	*(u32*)((u8*)pZoran->pFrameBuffer+
		pZoran->TabSlotOffset[pZoran->CapSlot]+
		pZoran->CapScanLineSize*((pZoran->CapHeight>>1)+1))=ZFS_NOTFILLED;
}

/**********************************************************************************************************************/
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/

static inline	void		ZoranIrq3ContinousSetupNewFrame(ZORAN * pZoran)
{
	ZR36120 *		pZR		=(ZR36120*)pZoran->pZR36120;

	pZR->StrideGrabControl.FrameGrab	=1;
}

/**********************************************************************************************************************/
		/****************************************************************/
		/*	printk in interupt handler in REALLY dangerous!		*/
		/****************************************************************/
		/*		interupts are disabled by the caller		*/

	void	ZoranIrqProc(int Irq,void * DeviceId,struct pt_regs * fake)
{
	ZORAN *			pZoran;
	ZR36120 *		pZR;
	int			IndexPCI;

	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++)
		{
		pZoran=TabZoran[IndexPCI];
		if (pZoran==NULL) continue;
		if (pZoran->IndexPCI!=IndexPCI) continue;
		pZR=(ZR36120*)pZoran->pZR36120;
		if (GIRQ==0)	{ if (pZR->IRQStatus.GIRQ0==0)
					{ /*printk("card did not irq\n");*/ continue; } }
		else		{ if (pZR->IRQStatus.GIRQ1==0)
					{ /*printk("card did not irq\n");*/ continue; } }

		switch(pZoran->ZRCapMode)
			{
			case ZCM_3CONTINOUSHACK:
				if (ZoranIrq3ContinousHackIsFrameOK(pZoran)==FALSE) break;
				ZoranIrq3ContinousSetNewFrame(pZoran);
				ZoranIrq3ContinousHackSetupNewFrame(pZoran);
				break;
			case ZCM_3CONTINOUS:
				if (ZoranIrq3ContinousIsFrameOK(pZoran)==FALSE) break;
				ZoranIrq3ContinousSetNewFrame(pZoran);
				ZoranIrq3ContinousSetupNewFrame(pZoran);
				break;
			case ZCM_CONTINOUS:
			case ZCM_GRAB:
			case ZCM_STOP:
			default:
				break;
			}
		if (GIRQ==0)	pZR->IRQStatus.GIRQ0=1;		// clear irq0
		else		pZR->IRQStatus.GIRQ1=1;		// clear irq1
		pZR->StrideGrabControl.VidOvf		=1;	// clear VidOvf
		}
}

