/*
   ZoranProc.c

   ZORAN ZR36120 module driver.
   Copyright (C) 1997  Vincent Labie (try ivsparis@pratique.fr with
				      for Vincent Labie in the suject)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	"Zoran.h"
#include	"ZoranProc.h"

#include	<linux/proc_fs.h>

#define			ZORAN_PROC_DIR_ENTRY_NAME_LEN	16
typedef	int (PROC_GET_INFO) (char * BufferIn,char ** start,off_t off,int BufferLenghtIn,int unused);

extern	int				ZoranMajor;

extern	ZORAN *				TabZoran[NB_MAX_ZORAN_CARD];

static	BOOL				bRegistred=FALSE;
static	struct	proc_dir_entry *	TabpZoranProcEntry[NB_MAX_ZORAN_CARD];
static	BOOL				TabbRegistred[NB_MAX_ZORAN_CARD];
static	PROC_GET_INFO *			TabProcGetInfo[NB_MAX_ZORAN_CARD];

#ifdef	CONFIG_PROC_FS
/**********************************************************************************************************************/
	int	ZoranReadProc(char * BufferIn,int BufferLenght,int IndexPCI)
{
	ZORAN *			pZoran;
	int			l;
	char *			Buffer=BufferIn;
	BOOL			bFound=FALSE;

	l=sprintf(Buffer,"ZR36120 driver version %i.%i, major %i\n",ZORAN_VERSION,ZORAN_SUBVERSION,ZoranMajor);
	Buffer+=l; BufferLenght-=l; if (BufferLenght<128) goto end;

	if ((IndexPCI>=0)&&(IndexPCI<NB_MAX_ZORAN_CARD))
		{
		pZoran=TabZoran[IndexPCI];
		if (pZoran!=NULL)
			{
			l=ZoranDumpRegBuf(pZoran,Buffer,BufferLenght);
			Buffer+=l; BufferLenght-=l;
			bFound=TRUE;
			}
		}
	if (bFound) goto end;
	l=sprintf(Buffer,"no ZR36120 found.\n");
	Buffer+=l; BufferLenght-=l; if (BufferLenght<128) goto end;
end:
	return(Buffer-BufferIn);
}

#if	(NB_MAX_ZORAN_CARD>4)
#error	Not enought ZoranReadProc functions defined
#endif

/**********************************************************************************************************************/
static	int	ZoranReadProc0(char * Buffer,char ** start,off_t off,int BufferLenght,int unused)
{	return(ZoranReadProc(Buffer,BufferLenght,0));	}
/**********************************************************************************************************************/
static	int	ZoranReadProc1(char * Buffer,char ** start,off_t off,int BufferLenght,int unused)
{	return(ZoranReadProc(Buffer,BufferLenght,1));	}
/**********************************************************************************************************************/
static	int	ZoranReadProc2(char * Buffer,char ** start,off_t off,int BufferLenght,int unused)
{	return(ZoranReadProc(Buffer,BufferLenght,2));	}
/**********************************************************************************************************************/
static	int	ZoranReadProc3(char * Buffer,char ** start,off_t off,int BufferLenght,int unused)
{	return(ZoranReadProc(Buffer,BufferLenght,3));	}
#endif	//CONFIG_PROC_FS

/**********************************************************************************************************************/
	int	ZoranProcRegister(void)
{
#ifdef	CONFIG_PROC_FS
	int			IndexPCI;

	memset(TabpZoranProcEntry,0,sizeof(TabpZoranProcEntry));
	memset(TabbRegistred,0,sizeof(TabbRegistred));
	memset(TabProcGetInfo,0,sizeof(TabProcGetInfo));
	TabProcGetInfo[0]=ZoranReadProc0;	TabProcGetInfo[1]=ZoranReadProc1;
	TabProcGetInfo[2]=ZoranReadProc2;	TabProcGetInfo[3]=ZoranReadProc3;
	bRegistred=TRUE;

	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++)
		{
		ZORAN *			pZoran=TabZoran[IndexPCI];
		if (pZoran!=NULL)
			{
			char *			szName;
			struct proc_dir_entry *	pProcEntry;
			if (TabProcGetInfo[IndexPCI]==NULL)
				{ ERR(("Zoran%d: warning no ZoranReadProc%d function.\n",IndexPCI,IndexPCI));
				continue; }

			szName=(char*)kmalloc(ZORAN_PROC_DIR_ENTRY_NAME_LEN,GFP_KERNEL);
			if (szName==NULL)
				{ ERR(("Zoran%d: can't kmalloc proc dir entry name.\n",IndexPCI));
				return(-ENOMEM); }
			sprintf(szName,"zoran%d",IndexPCI);

			TabpZoranProcEntry[IndexPCI]=(struct proc_dir_entry*)
							kmalloc(sizeof(struct proc_dir_entry),GFP_KERNEL);
			if (TabpZoranProcEntry[IndexPCI]==NULL)
				{ ERR(("Zoran%d: can't kmalloc proc dir entry.\n",IndexPCI));
				kfree(szName); return(-ENOMEM); }
			pProcEntry=TabpZoranProcEntry[IndexPCI];
			memset(pProcEntry,0,sizeof(struct proc_dir_entry));
			pProcEntry->low_ino	=0;				/* low_ino: the inode -- dynamic */
			pProcEntry->namelen	=strlen(szName);		/* len of name */
			pProcEntry->name	=szName;			/* name */
			pProcEntry->mode	=S_IFREG|S_IRUGO;		/* mode: regular, read by anyone */
			pProcEntry->nlink	=1;				/* nlinks */
			pProcEntry->uid		=0;				/* owner (root) */
			pProcEntry->gid		=0;				/* group (root) */
			pProcEntry->size	=0;				/* size is 0 */
			pProcEntry->ops		=NULL;				/* inode ops unused */
			pProcEntry->get_info	=TabProcGetInfo[IndexPCI];	/* use "infos" instead*/
			
			if (proc_register_dynamic(&proc_root,TabpZoranProcEntry[IndexPCI])!=0)
				{ ERR(("can't register Zoran%d.\n",IndexPCI)); return(-ENOSYS); }
			TabbRegistred[IndexPCI]=TRUE;
			}
		}
	register_symtab(NULL);
#endif	//CONFIG_PROC_FS
	return(0);
}

/**********************************************************************************************************************/
	void	ZoranProcUnregister(void)
{
#ifdef	CONFIG_PROC_FS
	int			IndexPCI;

	if (bRegistred==FALSE) return;	// hasn't even pass trough ZoranProcRegister

	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++)
		if (TabpZoranProcEntry[IndexPCI]!=NULL)
			{
			if (TabbRegistred[IndexPCI])
				proc_unregister(&proc_root,TabpZoranProcEntry[IndexPCI]->low_ino);
			if (TabpZoranProcEntry[IndexPCI]->name!=NULL)
				kfree((char*)TabpZoranProcEntry[IndexPCI]->name);
			kfree(TabpZoranProcEntry[IndexPCI]);
			TabpZoranProcEntry[IndexPCI]=NULL;
			}
#endif	//CONFIG_PROC_FS
}

